
INFO = {
  "dtiplayground": { 
    "version" : "0.3.7"
  },
  "dmriprep": {
    "version" : "0.3.7"

  },
  "dtiatlasbuilder": {
    "version" : "0.1.0"
  },
  "autotract" : {
    "version" : "0.0.1"
  }
}
